class InvalidRoute(ValueError):
    """
    Defines an invalid route definition condition.
    """


class InvalidConnection(ValueError):
    """
    Defines an invalid connection definition condition.
    """
